# -*- coding: utf-8 -*-

import json

import requests
from django.utils.timezone import now
from django.conf import settings

from . import fragments


class QF(object):
    def __init__(self):
        self.url = settings.ZOO_NANTERRE_APPLICATIONS['implicit']['url']

    def request(self, service, fragments):
        msg = {
            'metadonnees': {
                'origine': 'agent',
                'service': service,
                'date-soumission': now().isoformat(' ').split('.')[0],
            },
            'fragments': fragments,
        }
        try:
            response = requests.post(
                self.url, data=json.dumps(msg),
                headers={
                    'Content-Type': 'application/json'
                })
        except requests.RequestException as e:
            return None, e
        try:
            return response.json(), None
        except ValueError as e:
            return None, e

    def lire_quotients_valides(self, date_de_reference):
        return self.request('lire-quotients-valides', [{
            'id': 1,
            'type': 'donnees-lecture-quotients-valides',
            'fragment': {
                'date-reference': date_de_reference.isoformat(),
            }
        }])

    def simuler_qf(self,
                   nature_qf,
                   annee_imposition,
                   rfr,
                   nb_parts,
                   monoparentalite,
                   annee_imposition_concubin=None,
                   rfr_concubin=None,
                   nb_parts_concubin=None):
        fragment = {
            'id': 1,
            'type': 'donnees-quotient-familial',
            'fragment': {
                'nature-qf': nature_qf,
                'annee-imposition': annee_imposition,
                'rfr': rfr,
                'nb-parts': nb_parts,
                'monoparentalite': monoparentalite,
                'annee-imposition-concubin': annee_imposition_concubin,
                'rfr-concubin': rfr_concubin,
                'nb-parts-concubin': nb_parts_concubin,
            }
        }
        response, error = self.request('simuler-quotient-familial', [fragment])
        if error:
            return None, error
        if isinstance(response, list):
            return response[0], None
        else:
            return None, response

    def calcul_qf(self,
                  individu,
                  nature_qf,
                  annee_imposition,
                  rfr,
                  nb_parts,
                  monoparentalite,
                  annee_imposition_concubin=None,
                  rfr_concubin=None,
                  nb_parts_concubin=None):
        parts = []
        federation = individu.content['cles_de_federation'].get('implicit')
        if not federation:
            parts.append({
                'id': 1,
                'type': 'nouvel-adulte',
                'fragment': fragments.individu_to_fragment(individu),
            })
            reference_key = 'beneficiaire'
            reference = '@1'
        else:
            reference_key = 'id-metier'
            reference = federation
        fragment = {
            'id': len(parts) + 1,
            'type': 'donnees-quotient-familial',
            'fragment': {
                reference_key: reference,
                'nature-qf': nature_qf,
                'annee-imposition': annee_imposition,
                'rfr': rfr,
                'nb-parts': nb_parts,
                'monoparentalite': monoparentalite,
                'annee-imposition-concubin': annee_imposition_concubin,
                'rfr-concubin': rfr_concubin,
                'nb-parts-concubin': nb_parts_concubin,
            }
        }
        parts.append(fragment)
        response, error = self.request('calcul-quotient-familial', parts)
        if error:
            return None, error
        if isinstance(response, list):
            if not federation:
                federation = response[0]['id-metier']
                individu.content['cles_de_federation']['implicit'] = federation
                individu.save()
                return response[1], None
            else:
                return response[0], None
        else:
            return None, response

    def lire_quotient_familial(self, individu, date_de_reference):
        federation = individu.content['cles_de_federation'].get('implicit')
        if not federation:
            return None, 'pas de clé métier'
        fragment = {
            'id': 1,
            'type': 'donnees-lecture-quotient-familial',
            'fragment': {
                'id-metier': federation,
                'date-reference': date_de_reference.isoformat(),
            }
        }
        response, error = self.request('lire-quotient-familial', [fragment])
        if error:
            return None, error
        if isinstance(response, list):
            return response, None
        else:
            return None, response

    def editer_carte(self, individu, id_qf):
        federation = individu.content['cles_de_federation'].get('implicit')
        if not federation:
            return None, 'pas de clé métier'
        fragment = {
            'id': 1,
            'type': 'donnees-carte-quotient-familial',
            'fragment': {
                'id-metier': federation,
                'id-qf': id_qf,
            }
        }
        response, error = self.request('editer-carte-quotient-familial', [fragment])
        if error:
            return None, error
        if isinstance(response, dict) and 'lien' in response:
            response = requests.get(response['lien'])
            if response.status_code == 200:
                return response.content, None
            return None, {
                'status_code': 'erreur-de-recuperation-du-lien',
                'lien': response['lien']
            }
        else:
            return None, response
