# -*- coding: utf-8 -*-
import urlparse
import datetime
import decimal
import xml.etree.ElementTree as ET

import requests

from collections import namedtuple

# Some POPOs!

Facture = namedtuple('Facture', ['date_facture', 'date_limite_recouvrement', 'etat',
                                 'incident_paiement', 'montant_initial', 'num', 'reste_a_payer',
                                 'creances', 'commentaire'])

Creance = namedtuple('Creance', [
    'imputation',
    'libelle',
    'montant',
    'num_creance',
    'commentaire',
    'champs_dynamiques'])


class Saga(object):
    def __init__(self, url, ns=None):
        self.url = url
        self.ns = ns or 'ns'

    @property
    def creance_url(self):
        return urlparse.urljoin(
            self.url,
            '/saga_web_testrsu/services/etat_facture_creance_literal')

    @property
    def paiement_url(self):
        return urlparse.urljoin(
            self.url,
            '/saga_web_testrsu/services/paiement_internet_ws_literal')

    def soap_call(self, url, body, content_tag, **kwargs):
        wrapper = '''<soap11:Envelope xmlns:soap11="http://schemas.xmlsoap.org/soap/envelope/"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:tns1="http://172.26.53.9:8081/saga_web_testrsu/services/paiement_internet_ws_literal"
    xmlns:tns0="http://172.26.53.9:8081/saga_web_testrsu/services/etat_facture_creance_literal"
    xmlns="ns">
      <soap11:Body>
         {body}
      </soap11:Body>
    </soap11:Envelope>'''
        body = body.format(**kwargs)
        body = wrapper.format(body=body)
        try:
            response = requests.post(url, body,
                                     headers={
                                         'Content-Type': 'application/xml',
                                         'SOAPAction': "'\"'\"",
                                     })
        except requests.RequestException as e:
            return None, e
        et = ET.fromstring(response.text)
        content_node = et.find('.//{%s}%s' % (self.ns, content_tag))
        if content_node is None:
            return None, 'no content node'
        # pluging XML inside XML is so stupid :(
        tree = ET.fromstring(content_node.text.encode('latin1'))

        if tree.tag == 'erreur':
            return None, tree.text

        return tree, None

    def resolve_code_tiers(self, code_tiers):
        body = '''
        <codeTiersFedere>
          <num_tiers>{code_tiers}</num_tiers>
        </codeTiersFedere>'''
        tree, error = self.soap_call(self.creance_url, body, 'codeTiersFedereReturn', ns=self.ns,
                                     code_tiers=code_tiers)
        if tree is None:
            return None, error

        if tree.tag != 'code_tiers_federe':
            return None, 'no code_tiers_federe node'
        return tree.text, None

    def get_child_content(self, tree, child_name):
        t = tree.find(child_name)
        if t:
            return t.text

    def factures(self, federation):
        body = '''
        <etatFactureParTiersFedere>
          <num_tiers>{federation}</num_tiers>
          <num_service>67</num_service>
          <type_facture>toute</type_facture>
          <periode_debut></periode_debut>
          <periode_fin></periode_fin>
          <detail>oui</detail>
        </etatFactureParTiersFedere>'''

        tree, error = self.soap_call(
            self.creance_url, body, 'etatFactureParTiersFedereReturn',
            federation=federation, ns=self.ns)
        if tree is None:
            return None, error

        def helper():
            for t in tree.findall('.//facture'):
                a = t.attrib

                def helper2():
                    for c in t.findall('.//creance'):
                        a = c.attrib
                        champs_dynamiques = {}
                        for champ_dynamique in c.findall('.//champ_dynamique'):
                            identifiant = champ_dynamique.attrib['identifiant']
                            valeur = champ_dynamique.attrib['valeur']
                            champs_dynamiques[identifiant] = valeur
                        yield Creance(
                            imputation=a['imputation'],
                            libelle=a['libelle'],
                            montant=decimal.Decimal(a['montant']),
                            num_creance=a['num_creance'],
                            commentaire=self.get_child_content(c, 'commentaire'),
                            champs_dynamiques=champs_dynamiques)

                def parse_date(d):
                    return datetime.datetime.strptime(d, '%d/%m/%Y').date()

                facture = Facture(
                    date_facture=parse_date(a['date_facture']),
                    date_limite_recouvrement=parse_date(a['date_limite_recouvrement']),
                    etat=a['etat'],
                    incident_paiement=a['incident_paiement'] != 'non',
                    montant_initial=decimal.Decimal(a['montant_initial']),
                    num=a['num'],
                    reste_a_payer=decimal.Decimal(a['reste_a_payer']),
                    commentaire=self.get_child_content(t, 'commentaire'),
                    creances=list(helper2()))

                yield facture
        return list(helper()), None

    def transaction(self, facture, urlretour_asynchrone, urlretour_synchrone, email):
        body = '''
        <Transaction>
          <num_service>67</num_service>
          <id_facture>{id_facture}</id_facture>
          <montant>{montant}</montant>
          <urlretour_asynchrone>{urlretour_asynchrone}</urlretour_asynchrone>
          <email>{email}</email>
          <urlretour_synchrone>{urlretour_synchrone}</urlretour_synchrone>
        </Transaction>'''
        tree, error = self.soap_call(
            self.paiement_url, body, 'TransactionReturn',
            id_facture=facture.num,
            montant=facture.reste_a_payer,
            urlretour_asynchrone=urlretour_asynchrone,
            urlretour_synchrone=urlretour_synchrone,
            email=email)
        if tree is None:
            return None, error

        if tree.tag != 'url':
            return None, 'tag is not error'

        return tree.text, None

    def page_retour_asynchrone(self, idop):
        body = '''
        <PageRetourAsynchrone>
          <idop>{idop}</idop>
        </PageRetourAsynchrone>'''
        tree, error = self.soap_call(
            self.paiement_url, body, 'PageRetourAsynchroneReturn',
            idop=idop)
        if tree is None:
            return None, error

        if tree.tag != 'ok':
            return None, 'tag is not error'
        result = {
            'code_tiers': tree.attrib.get('code_tiers'),
            'etat': tree.attrib['etat'],
            'email': tree.attrib.get('email'),
            'num_service': tree.attrib.get('num_service'),
        }
        return result, None

    def page_retour_synchrone(self, idop):
        body = '''
        <PageRetourSynchrone>
          <idop>{idop}</idop>
        </PageRetourSynchrone>'''
        tree, error = self.soap_call(
            self.paiement_url, body, 'PageRetourSynchroneReturn',
            idop=idop)
        if tree is None:
            return None, error

        if tree.tag != 'ok':
            return None, 'tag is not ok'
        result = {
            'code_tiers': tree.attrib.get('code_tiers'),
            'etat': tree.attrib['etat'],
            'email': tree.attrib.get('email'),
            'num_service': tree.attrib.get('num_service'),
        }
        return result, None
