# -*- coding: utf-8 -*-
#
# zoo - data management system
# Copyright (C) 2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import isodate
import copy
import re
import traceback

from django.conf import settings
from django.shortcuts import get_object_or_404
from django.db.models.query import Q
from django.db.transaction import non_atomic_requests, atomic
from django.core.urlresolvers import reverse
from django.http import Http404, HttpResponse
from django.utils.timezone import now

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import serializers

from zoo.zoo_meta.models import EntitySchema, RelationSchema
from zoo.zoo_data.models import Entity, Relation, Transaction, Log

from . import utils, fragments, saga, qf


def flatten_errors(serializer_errors):
    errors = []
    for k, v in serializer_errors.items():
        errors.append(u'%s: %s' % (k, v))
    return errors


def individu_to_text(individu):
    d = individu.content
    text = d['nom_de_naissance'] + ' '
    if d.get('nom_d_usage'):
        text += '(' + d['nom_d_usage'] + ') '
    text += d['prenoms']
    date = isodate.parse_date(d['date_de_naissance'])
    text += ' - %02d/%02d/%04d - ' % (date.day, date.month, date.year)
    text += d['genre']
    if d.get('statut_legal'):
        text += '/' + d['statut_legal']
    return text


def individu_to_response(individu, add_text=False, add_conjoint=True, add_enfant=True,
                         add_parents=True):
    '''Serialize a person'''
    d = individu.content.copy()
    d['id'] = individu.id

    if hasattr(individu, 'age_label'):
        d['age_label'] = individu.age_label
    if hasattr(individu, 'age'):
        d['age'] = individu.age
    if hasattr(individu, 'similarity'):
        d['score'] = individu.similarity
    if hasattr(individu, 'adresses'):
        d['adresses'] = individu.adresses
    if hasattr(individu, 'responsabilite_legale'):
        d['responsabilite_legale'] = individu.responsabilite_legale
    if add_enfant and hasattr(individu, 'enfants'):
        d['enfants'] = [individu_to_response(enfant) for enfant in individu.enfants]
    if add_parents and hasattr(individu, 'parents'):
        d['parents'] = [individu_to_response(parent) for parent in individu.parents]
    if add_conjoint and hasattr(individu, 'union'):
        d[utils.UNION_REL] = individu_to_response(individu.union)
        d['union_statut'] = individu.union_statut
    if add_text:
        d['text'] = individu_to_text(individu)
    d['date_de_creation'] = individu.created.created.isoformat()
    return d


class TransactionalView(APIView):
    def dispatch(self, request, *args, **kwargs):
        if request.method not in ['GET', 'HEAD', 'OPTIONS']:
            with atomic():
                return super(TransactionalView, self).dispatch(request, *args, **kwargs)
        else:
                return super(TransactionalView, self).dispatch(request, *args, **kwargs)

    def initial(self, request, *args, **kwargs):
        if request.method not in ['GET', 'HEAD', 'OPTIONS']:
            self.transaction = Transaction.get_transaction()
            self.transaction.content = {}
        super(TransactionalView, self).initial(request, *args, **kwargs)

    def handle_exception(self, exc):
        if hasattr(self, 'transaction'):
            content = {
                'request': self.request.data,
                'status_code': 500,
                '$exc_detail': unicode(exc),
                '$exc_tb': traceback.format_exc(),
            }
            self.transaction.content = content
            self.transaction.save()
        return super(TransactionalView, self).handle_exception(exc)

    def finalize_response(self, request, response, *args, **kwargs):
        if hasattr(self, 'transaction') and not getattr(response, 'exception', False):
            content = {
                'url': request.build_absolute_uri(),
                'request': request.data,
                'response': response.data,
                'status_code': response.status_code,
            }
            self.transaction.content.update(content)
            self.transaction.save()
        return super(TransactionalView, self).finalize_response(request, response, *args, **kwargs)


class SearchView(APIView):
    def get(self, request, format=None):
        try:
            limit = int(request.GET.get('limit', ''))
        except ValueError:
            limit = 100
        try:
            offset = int(request.GET.get('offset', ''))
        except ValueError:
            offset = 0
        try:
            threshold = float(request.GET.get('threshold', ''))
        except ValueError:
            threshold = 0.2
        search = utils.PersonSearch(limit=threshold)
        if 'q' in request.GET:
            search = search.search_query(request.GET['q'])
        else:
            prenom = request.GET.get('prenom')
            nom = request.GET.get('nom')
            date_de_naissance = request.GET.get('date_de_naissance')
            cle = request.GET.get('cle')
            email = request.GET.get('email', '').strip()
            name_id = request.GET.get('NameID')
            if prenom or nom:
                search = search.search_name(prenom, nom)
            if date_de_naissance and search.match_birthdate(date_de_naissance):
                search = search.search_birthdate(date_de_naissance)
            if cle:
                search = search.search_identifier(cle)
            if email:
                search = search.search_email(email)
            if name_id:
                search = search.search_identifier(name_id, key='authentic')
            for key in request.GET:
                if key.startswith('cle_'):
                    cle = request.GET[key]
                    search = search.search_identifier(cle, key=key[4:])
        if 'statut_legal' in request.GET:
            search.search_statut_legal(request.GET['statut_legal'])
        data = [individu_to_response(person, add_text=True)
                for person in search[offset:offset + limit]]
        return Response({
            'err': 0,
            'offset': offset,
            'limit': limit,
            'count': len(data),
            'data': data,
            'meta': {
                'applications': utils.PersonSearch.applications(),
            }
        })

search = non_atomic_requests(SearchView.as_view())


class IndividuViewMixin(object):
    def get_individu(self, identifier, **kwargs):
        qs = Entity.objects.prefetch_related(
            'left_relations__schema', 'left_relations__right',
            'right_relations__schema', 'right_relations__left',
        )
        try:
            identifier = int(identifier)
        except ValueError:
            return get_object_or_404(qs, schema__slug=utils.INDIVIDU_ENT,
                                     content__cles_de_federation__authentic=identifier,
                                     **kwargs)
        else:
            return get_object_or_404(qs, schema__slug=utils.INDIVIDU_ENT, id=identifier, **kwargs)


class ReseauView(IndividuViewMixin, TransactionalView):
    def get(self, request, identifier, format=None):
        individu = self.get_individu(identifier)
        utils.PersonSearch.decorate_individu(individu)
        return Response({
            'err': 0,
            'data': individu_to_response(individu),
            'meta': {
                'applications': utils.PersonSearch.applications(),
            }
        })

    @atomic
    def post(self, request, identifier, format=None):
        individu = self.get_individu(identifier)
        utils.PersonSearch.decorate_individu(individu)
        serializer = CreateIndividuSerializer(data=request.data, partial=True)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)
        if (serializer.validated_data.get(utils.ADRESSE_ENT)
                and individu.content['statut_legal'] == 'mineur'):
            return Response({
                'err': 1,
                'errors': [
                    u'l\'adresse est interdite pour un mineur',
                ],
            }, status=400)
        if serializer.validated_data.get('cles_de_federation'):
            errors = []
            for key, value in serializer.validated_data.get('cles_de_federation').items():
                if value:
                    for other in Entity.objects.exclude(id=individu.id).filter(
                            **{'content__cles_de_federation__%s' % key: value}):
                        errors.append(u'la clé %s %s est déjà utilisée par l\'individu #%s' % (
                            key, value, other.id))
            if errors:
                return Response({
                    'err': 1,
                    'errors': errors,
                }, status=400)

        transaction = self.transaction
        v = serializer.validated_data
        individu.modified = transaction
        identite_modifie = False
        contact_modifie = False
        if 'prenoms' in v:
            individu.content['prenoms'] = v['prenoms'].upper()
            identite_modifie = True
        if 'nom_de_naissance' in v:
            individu.content['nom_de_naissance'] = v['nom_de_naissance'].upper()
            identite_modifie = True
        if 'nom_d_usage' in v:
            individu.content['nom_d_usage'] = v['nom_d_usage'].upper()
            identite_modifie = True
        if 'date_de_naissance' in v:
            individu.content['date_de_naissance'] = v['date_de_naissance'].isoformat()
            identite_modifie = True
        if 'genre' in v:
            individu.content['genre'] = v['genre']
            identite_modifie = True
        if 'cles_de_federation' in v:
            cles_de_federation = individu.content.setdefault('cles_de_federation', {})
            for name in settings.ZOO_NANTERRE_APPLICATIONS:
                if v['cles_de_federation'] and name in v['cles_de_federation']:
                    key = v['cles_de_federation'][name]
                    if not key: # remove key from individu
                        if name in cles_de_federation:
                            del cles_de_federation[name]
                    else:
                        cles_de_federation[name] = key
        if 'email' in v:
            individu.content['email'] = v['email']
            contact_modifie = True
        if 'telephones' in v:
            individu.content['telephones'] = v['telephones']
            contact_modifie = True
        individu.save()
        if identite_modifie:
            fragments.MiseAJourIdentite.pour_chaque_application(
                [individu],
                meta=serializer.journal_meta,
                transaction=transaction)
        if contact_modifie:
            fragments.MiseAJourInformationsContact.pour_chaque_application(
                [individu],
                meta=serializer.journal_meta,
                transaction=transaction)
        if individu.content['statut_legal'] == 'majeur' and utils.ADRESSE_ENT in v:
            adresse = utils.adresse(individu)
            utils.upper_dict(v['adresse'])
            adresse.content = v['adresse']
            adresse.modified = individu.modified
            adresse.save()
            habitants = Entity.objects.filter(left_relations__right=adresse)
            fragments.SignalementChangementAdresse.pour_chaque_application(
                habitants,
                meta=serializer.journal_meta,
                transaction=transaction)
        # no need to update children and husband/wife adresses,
        # they sould already have the same address
        utils.journalize(
            individu,
            meta=serializer.journal_meta,
            transaction=transaction,
            text=u'Mise à jour des informations')
        return Response({
            'err': 0,
            'data': individu_to_response(individu),
        })

reseau = non_atomic_requests(ReseauView.as_view())


class ReseauListView(IndividuViewMixin, APIView):
    def get(self, request, identifier, format=None):
        # permet de ne voir que le conjoint ou que les enfants
        conjoint = request.GET.get('conjoint') is not None
        enfants = request.GET.get('enfants') is not None
        foyer = request.GET.get('foyer') is not None

        # les deux options ensemble s'annulent
        if int(conjoint) + int(enfants) + int(foyer) > 1:
            conjoint = False
            enfants = False
            foyer = False

        individu = self.get_individu(identifier)
        utils.PersonSearch.decorate_individu(individu)
        data = []
        # on ajoute l'individu visé que si on n'a pas demandé à ne voir que le conjoint ou que les
        # enfants
        if not conjoint and not enfants:
            data.append(individu_to_response(individu, add_parents=False, add_enfant=False,
                                             add_conjoint=False, add_text=True))
        if hasattr(individu, 'union') and not enfants:
            data.append(individu_to_response(individu.union, add_text=True))

        if hasattr(individu, 'enfants') and not conjoint:
            enfants = set(individu.enfants)
            if foyer and hasattr(individu, 'union'):
                utils.PersonSearch.decorate_individu(individu.union)
                enfants.update(individu.union.enfants)

            for enfant in enfants:
                utils.PersonSearch.add_age(enfant)
            # ordonne les enfants du plus agé au plus jeune
            for enfant in sorted(enfants, key=lambda e: e.age, reverse=True):
                data.append(individu_to_response(enfant, add_text=True))

        return Response({
            'err': 0,
            'data': data,
        })

reseau_liste = ReseauListView.as_view()


class JournalSerializerMixin(serializers.Serializer):
    def __init__(self, instance=None, data=None, **kwargs):
        meta = self.journal_meta = {}
        if data:
            data = data.copy()
            for key in data:
                if key.startswith('journal_'):
                    meta[key[8:]] = data[key]
        super(JournalSerializerMixin, self).__init__(instance=instance, data=data, **kwargs)


class AdresseSerializer(serializers.Serializer):
    at = serializers.CharField(allow_blank=True)
    streetnumber = serializers.CharField(allow_blank=True)
    streetnumberext = serializers.CharField(allow_blank=True)
    streetname = serializers.CharField()
    ext1 = serializers.CharField(allow_blank=True)
    ext2 = serializers.CharField(allow_blank=True)
    streetmatriculation = serializers.CharField(allow_blank=True)
    zipcode = serializers.CharField(allow_blank=True)
    inseecode = serializers.CharField(allow_blank=True)
    city = serializers.CharField()
    country = serializers.CharField()
    adresse_inconnnue = serializers.BooleanField(default=False, required=False)


class TelephoneSerializer(serializers.Serializer):
    numero = serializers.RegexField('^[0-9 .-]*$')
    type = serializers.ChoiceField(
        choices=[
            'maison',
            'mobile',
            'pro',
            'autre',
        ])


class CreateIndividuSerializer(JournalSerializerMixin):
    genre = serializers.ChoiceField(
        choices=[
            'femme',
            'homme',
            'autre',
        ])
    prenoms = serializers.CharField(max_length=128)
    nom_d_usage = serializers.CharField(max_length=128, allow_blank=True)
    nom_de_naissance = serializers.CharField(max_length=128)
    date_de_naissance = serializers.DateField()
    email = serializers.EmailField(allow_blank=True)
    adresse = AdresseSerializer()
    telephones = TelephoneSerializer(many=True)
    cles_de_federation = serializers.DictField(child=serializers.CharField(allow_blank=True), required=False,
                                               default=None)


class CreateIndividu(TransactionalView):
    def post(self, request):
        serializer = CreateIndividuSerializer(data=request.data)
        transaction = self.transaction
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)
        if serializer.validated_data.get('cles_de_federation'):
            return Response({
                'err': 1,
                'errors': [
                    'clés de fédération non admises lors de la création d\'un individu'
                ],
            }, status=400)
        schema = EntitySchema.objects.get(slug=utils.INDIVIDU_ENT)
        schema_adresse = EntitySchema.objects.get(slug=utils.ADRESSE_ENT)
        habite_schema = RelationSchema.objects.get(slug=utils.HABITE_REL)
        v = serializer.validated_data
        v['adresse_inconnnue'] = False  # new individual always have known adresses
        cles_de_federation = {}
        individu = Entity(
            schema=schema,
            created=transaction,
            content={
                'prenoms': v['prenoms'].upper(),
                'nom_d_usage': v['nom_d_usage'].upper(),
                'nom_de_naissance': v['nom_de_naissance'].upper(),
                'date_de_naissance': v['date_de_naissance'].isoformat(),
                'telephones': v['telephones'],
                'genre': v['genre'],
                'statut_legal': 'majeur',
                'cles_de_federation': cles_de_federation,
                'email': v['email'],
            })
        content = v[utils.ADRESSE_ENT].copy()
        utils.upper_dict(content)
        content['adresse_inconnnue'] = False
        individu.save()
        adresse = Entity.objects.create(
            schema=schema_adresse,
            created=individu.created,
            content=content,
        )
        Relation.objects.create(
            created=individu.created,
            left=individu,
            right=adresse,
            schema=habite_schema,
            content={
                'principale': False,
            }
        )
        individu.adresses = [adresse.content]
        utils.journalize(
            individu,
            transaction=transaction,
            meta=serializer.journal_meta,
            text=u'Création de l\'individu')
        return Response({
            'err': 0,
            'data': individu_to_response(individu),
        })

create_individu = CreateIndividu.as_view()


class EnfantSerializer(serializers.Serializer):
    genre = serializers.ChoiceField(
        choices=[
            'femme',
            'homme',
            'autre',
        ])
    prenoms = serializers.CharField(max_length=128)
    nom_de_naissance = serializers.CharField(max_length=128)
    date_de_naissance = serializers.DateField()
    email = serializers.EmailField(allow_blank=True)
    telephones = TelephoneSerializer(many=True)


class DeclarationResponsabiliteLegaleSerializer(JournalSerializerMixin):
    statut = serializers.ChoiceField(
        choices=[
            'parent',
            'tiers_de_confiance',
            'representant_personne_morale_qualifiee',
        ])
    enfant_id = serializers.IntegerField(required=False)
    enfant = EnfantSerializer(required=False)


class DeclarationResponsabiliteLegale(IndividuViewMixin, TransactionalView):
    def post(self, request, identifier, format=None):
        individu = self.get_individu(identifier, content__statut_legal='majeur')
        serializer = DeclarationResponsabiliteLegaleSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)
        v = serializer.validated_data
        if v.get('enfant_id') and v.get('enfant'):
            return Response({
                'err': 1,
                'errors': [
                    u'vous ne devez pas fournir les paramètres enfant_id et enfant',
                ]
            }, status=400)
        if not v.get('enfant_id') and not v.get('enfant'):
            return Response({
                'err': 1,
                'errors': [
                    u'vous devez fournir un des paramètres enfant_id ou enfant',
                ]
            }, status=400)

        transaction = self.transaction
        individu_schema = EntitySchema.objects.get(slug=utils.INDIVIDU_ENT)
        responsabilite_legale_schema = RelationSchema.objects.get(
            slug=utils.RESPONSABILITE_LEGALE_REL)
        habite_schema = RelationSchema.objects.get(slug=utils.HABITE_REL)
        adresse = list(utils.adresses(individu))[0][0]

        if v.get('enfant_id'):  # existing child
            try:
                enfant = Entity.objects.get(
                    schema=individu_schema,
                    content__statut_legal='mineur',
                    id=v['enfant_id'])
            except Entity.DoesNotExist:
                return Response({
                    'err': 1,
                    'errors': [
                        u'enfant_id: identifiant inconnu',
                    ]
                }, status=400)
        else:  # new child
            enfant = Entity.objects.create(
                created=transaction,
                schema=individu_schema,
                content={
                    'prenoms': v['enfant']['prenoms'].upper(),
                    'nom_de_naissance': v['enfant']['nom_de_naissance'].upper(),
                    'nom_d_usage': '',
                    'email': v['enfant']['email'],
                    'date_de_naissance': v['enfant']['date_de_naissance'].isoformat(),
                    'genre': v['enfant']['genre'],
                    'telephones': v['enfant']['telephones'],
                    'statut_legal': 'mineur',
                    'cles_de_federation': {},
                })
        # verify child has not already two parents
        if v['statut'] == 'parent':
            if enfant.right_relations.filter(
                    content__statut='parent',
                    schema__slug=utils.RESPONSABILITE_LEGALE_REL).count() > 1:
                return Response({
                    'err': 1,
                    'errors': [
                        u'enfant_id: cet enfant a déjà deux parents',
                    ]
                }, status=400)

        # verify child is not already linked to this parent
        if enfant.right_relations.filter(schema__slug=utils.RESPONSABILITE_LEGALE_REL,
                                         left=individu).exists():
            return Response({
                'err': 1,
                'errors': [
                    'enfant_id: cet enfant a déjà cet adulte pour responsable légal',
                ]
            }, status=400)

        relation = enfant.right_relations.create(
            created=transaction,
            schema=responsabilite_legale_schema,
            content={
                'statut': v['statut'],
            },
            left=individu)

        # if the child does not already live at this address, link him to it
        if not enfant.left_relations.filter(schema=habite_schema, right=adresse):
            enfant.left_relations.create(
                created=transaction,
                schema=habite_schema,
                content={
                    'principale': False,
                },
                right=adresse)
        utils.journalize(
            individu, meta=serializer.journal_meta,
            transaction=transaction,
            text=u'Déclaration de responsabilité légale',
            enfant_id=enfant.id,
            enfant_text=individu_to_text(enfant))
        utils.journalize(
            enfant, meta=serializer.journal_meta,
            transaction=transaction,
            text=u'Déclaration de responsabilité légale',
            adulte_id=individu.id,
            adulte_text=individu_to_text(individu))
        fragments.DeclarationResponsabiliteLegaleEnfant.pour_chaque_application(
            relation,
            meta=serializer.journal_meta,
            transaction=transaction)
        return Response({
            'err': 0,
            'data': individu_to_response(enfant),
        })

declaration_responsabilite_legale = DeclarationResponsabiliteLegale.as_view()


class DeclarationUnionSerializer(JournalSerializerMixin):
    individu_id_1 = serializers.IntegerField()
    individu_id_2 = serializers.IntegerField()
    adresse_commune = serializers.ChoiceField(choices=[1, 2])
    statut = serializers.ChoiceField(
        choices=['pacs/mariage', 'unionlibre'])


class DeclarationUnion(TransactionalView):
    def post(self, request, format=None):
        errors = []
        serializer = DeclarationUnionSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)
        v = serializer.validated_data

        try:
            individu1 = Entity.objects.get(id=v['individu_id_1'],
                                           schema__slug=utils.INDIVIDU_ENT)
        except Entity.DoesNotExist:
            individu1 = None

        try:
            individu2 = Entity.objects.get(id=v['individu_id_2'],
                                           schema__slug=utils.INDIVIDU_ENT)
        except Entity.DoesNotExist:
            individu2 = None

        if not individu1:
            errors.append(u'individu_id_1: identifiant inconnu')
        if not individu2:
            errors.append(u'individu_id_2: identifiant inconnu')
        if errors:
            return Response({
                'err': 1,
                'errors': errors,
            })

        # vérification sur la majorité
        if (not individu1.content['statut_legal'] == 'majeur' or
                not individu2.content['statut_legal'] == 'majeur'):
            return Response({
                'err': 1,
                'errors': [
                    u'les deux individus ne sont pas tous les deux majeurs',
                ],
            })

        qs = Relation.objects.filter(schema__slug=utils.UNION_REL)
        already1 = qs.filter(Q(left=individu1) | Q(right=individu1)).exists()
        already2 = qs.filter(Q(left=individu2) | Q(right=individu2)).exists()

        if already1:
            errors.append(u'individu_id_1: cet individu est déjà dans une relation maritale')
        if already2:
            errors.append(u'individu_id_2: cet individu est déjà dans une relation maritale')
        if errors:
            return Response({
                'err': 1,
                'errors': errors,
            })

        transaction = self.transaction
        union_schema = RelationSchema.objects.get(slug=utils.UNION_REL)
        relation = Relation.objects.create(
            created=transaction,
            schema=union_schema,
            left=individu1,
            right=individu2,
            content={
                'statut': v['statut'],
            })

        adresse1 = utils.adresse(individu1)
        adresse2 = utils.adresse(individu2)
        assert adresse1 != adresse2, 'people should not have the same address'
        if v['adresse_commune'] == 1:
            qs = Entity.objects.filter(left_relations__right=adresse1)
            Relation.objects.exclude(left__in=qs).filter(right=adresse2) \
                    .update(right=adresse1, modified=transaction)
            adresse2.delete()
        else:
            qs = Entity.objects.filter(left_relations__right=adresse2)
            Relation.objects.exclude(left__in=qs).filter(right=adresse1) \
                    .update(right=adresse2, modified=transaction)
            adresse1.delete()

        utils.journalize(
            individu1, meta=serializer.journal_meta,
            text=u'Déclaration d\'union',
            transaction=transaction,
            left_id=individu1.id,
            left_text=individu_to_text(individu1),
            right_id=individu2.id,
            right_text=individu_to_text(individu2))
        utils.journalize(
            individu2, meta=serializer.journal_meta,
            text=u'Déclaration d\'union',
            transaction=transaction,
            left_id=individu1.id,
            left_text=individu_to_text(individu1),
            right_id=individu2.id,
            right_text=individu_to_text(individu2))
        fragments.DeclarationUnion.pour_chaque_application(
            relation,
            meta=serializer.journal_meta,
            transaction=transaction)
        return Response({
            'err': 0,
        })

declaration_union = DeclarationUnion.as_view()


class JournalView(IndividuViewMixin, APIView):
    FILTER_RE = re.compile('^[a-z_]*$')

    def get(self, request, identifier, format=None):
        individu = self.get_individu(identifier)
        filters = {}
        for key in request.GET:
            if key.startswith('filter_') and self.FILTER_RE.match(key):
                value = request.GET.getlist(key)
                if len(value) > 1:
                    filters['content__' + key[7:] + '__in'] = value
                else:
                    filters['content__' + key[7:]] = value[0]
        try:
            limit = int(request.GET.get('limit', ''))
        except ValueError:
            limit = 10
        try:
            cookie = int(request.GET.get('cookie', ''))
        except ValueError:
            cookie = 0
        limit = min(limit, 100)
        qs = Log.objects.filter(entity=individu)
        if filters:
            qs = qs.filter(**filters)
        if cookie:
            qs = qs.filter(id__lt=cookie)
        qs = qs[:limit + 1]
        data = [{'timestamp': log.timestamp.isoformat(), 'content': log.content}
                for log in qs[:limit]]
        content = {
            'err': 0,
            'data': data,
        }
        if len(qs) > limit:
            max_id = min(log.id for log in qs[:limit])
            more_url = request.build_absolute_uri(
                reverse('rsu-api-journal', kwargs={'identifier': individu.id}))
            more_url += '?limit=%d&cookie=%s' % (limit, max_id)
            content['more'] = more_url
        return Response(content)

    def post(self, request, identifier, format=None):
        qs = Entity.objects.prefetch_related(
            'left_relations__schema', 'left_relations__right',
            'right_relations__schema', 'right_relations__left',
        )
        individu = get_object_or_404(qs, schema__slug=utils.INDIVIDU_ENT, id=identifier)
        Log.objects.create(
            entity=individu,
            content=request.data)
        return Response({
            'err': 0,
        })

journal = JournalView.as_view()


class DeclarationAdressePrincipaleSerializer(JournalSerializerMixin):
    adresse_principale = serializers.IntegerField()


class DeclarationAdressePrincipaleView(IndividuViewMixin, TransactionalView):
    def post(self, request, identifier, format=None):
        individu = self.get_individu(identifier)

        serializer = DeclarationAdressePrincipaleSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)

        if individu.content['statut_legal'] != 'mineur':
            return Response({
                'err': 1,
                'errors': [
                    u'cet individu n\'est pas mineur'
                ]
            })
        adresses = list(utils.adresses(individu))
        idx = serializer.validated_data['adresse_principale']
        if not (1 <= idx <= len(adresses)):
            return Response({
                'err': 1,
                'errors': [
                    u'identifiant d\'adresse inconnu'
                ]
            })
        transaction = self.transaction
        for i, (adresse, rel) in enumerate(adresses):
            if i + 1 == idx:
                if not rel.content['principale']:
                    rel.content['principale'] = True
                    rel.modified = transaction
                    rel.save()
            else:
                if rel.content['principale']:
                    rel.content['principale'] = False
                    rel.modified = transaction
                    rel.save()
        fragments.SignalementChangementAdresse.pour_chaque_application(
            [individu],
            meta=serializer.journal_meta,
            transaction=transaction)
        utils.journalize(
            individu, meta=serializer.journal_meta,
            text=u'Déclaration d\'adresse principale',
            adresse_idx=idx,
            transaction=transaction)
        return Response({
            'err': 0
        })

declaration_adresse_principale = DeclarationAdressePrincipaleView.as_view()


class ChangementDeSituationMaritaleSerializer(JournalSerializerMixin):
    statut = serializers.ChoiceField(
        choices=['pacs/mariage', 'unionlibre'])


class ChangementDeSituationMaritaleView(IndividuViewMixin, TransactionalView):
    def post(self, request, identifier, format=None):
        individu = self.get_individu(identifier)

        if individu.content['statut_legal'] != 'majeur':
            return Response({
                'err': 1,
                'errors': [
                    u'cet individu n\'est pas majeur',
                ]
            })

        conjoint, conjoint_rel = utils.conjoint(individu)
        if not conjoint:
            return Response({
                'err': 1,
                'errors': [
                    u"cet individu n'a pas actuellement de relation maritale",
                ]
            })

        serializer = ChangementDeSituationMaritaleSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': serializer.errors,
            }, status=400)

        if conjoint_rel.content['statut'] == serializer.validated_data['statut']:
            return Response({
                'err': 1,
                'errors': [
                    u'la relation maritale est déjà de ce type'
                ]
            })

        old_statut = conjoint_rel.content['statut']
        conjoint_rel.content['statut'] = serializer.validated_data['statut']
        conjoint_rel.modified = self.transaction
        conjoint_rel.save()

        utils.journalize(
            individu,
            meta=serializer.journal_meta,
            text=u'Changement de situation maritale',
            old_statut=old_statut,
            statut=conjoint_rel.content['statut'],
            transaction=self.transaction)
        utils.journalize(
            conjoint,
            meta=serializer.journal_meta,
            text=u'Changement de situation maritale',
            old_statut=old_statut,
            statut=conjoint_rel.content['statut'],
            transaction=self.transaction)

        fragments.DeclarationUnion.pour_chaque_application(
            conjoint_rel,
            meta=serializer.journal_meta,
            transaction=self.transaction)

        return Response({
            'err': 0
        })

changement_de_situation_maritale = ChangementDeSituationMaritaleView.as_view()


class SeparationView(IndividuViewMixin, TransactionalView):
    def enfants_communs(self, individu, conjoint):
        children = set(enfant for enfant, rel in utils.enfants(individu))
        children &= set(enfant for enfant, rel in utils.enfants(conjoint))

        return sorted(children, key=lambda e: e.id)

    def get(self, request, identifier, format=None):
        individu = self.get_individu(identifier)

        if individu.content['statut_legal'] != 'majeur':
            return Response({
                'err': 1,
                'errors': [
                    u'cet individu n\'est pas majeur',
                ]
            })

        conjoint, conjoint_rel = utils.conjoint(individu)
        if not conjoint:
            return Response({
                'err': 1,
                'errors': [
                    u'cet individu n\'a pas actuellement de relation maritale',
                ]
            })
        return Response({
            'err': 0,
            'data': {
                utils.UNION_REL: individu_to_response(conjoint),
                'union_statut': conjoint_rel.content['statut'],
                'enfants': [individu_to_response(enfant, add_text=True) for enfant in
                            self.enfants_communs(individu, conjoint)],
            },
        })

    def post(self, request, identifier, format=None):
        individu = self.get_individu(identifier)

        serializer = JournalSerializerMixin(data=request.data)
        serializer.is_valid()

        utils.PersonSearch.decorate_individu(individu)

        if individu.content['statut_legal'] != 'majeur':
            return Response({
                'err': 1,
                'errors': [
                    u'cet individu n\'est pas majeur',
                ]
            })

        conjoint, conjoint_rel = utils.conjoint(individu)
        if not conjoint:
            return Response({
                'err': 1,
                'errors': [
                    u'cet individu n\'a pas actuellement de relation maritale',
                ]
            })

        enfants_communs = self.enfants_communs(individu, conjoint)

        adresse_1 = utils.adresse(individu)

        errors = []
        enfant_ids = [enfant.id for enfant in enfants_communs]
        adresse_principale_1 = request.data.get('adresse_principale_1', [])
        adresse_principale_2 = request.data.get('adresse_principale_2', [])
        if not isinstance(adresse_principale_1, list):
            errors.append(u'adresse_principale_1: doit être une liste d\'identifiants.')
        if not isinstance(adresse_principale_2, list):
            errors.append(u'adresse_principale_2: doit être une liste d\'identifiants.')

        for key in adresse_principale_1:
            if key not in enfant_ids:
                errors.append(u'adresse_principale_1: l\'enfant %s n\'est pas commun' % key)

        for key in adresse_principale_2:
            if key not in enfant_ids:
                errors.append(u'adresse_principale_2: l\'enfant %s n\'est pas commun' % key)

        if set(adresse_principale_1) & set(adresse_principale_2):
            errors.append(u'un enfant ne peut pas avoir deux adresses principales')

        if errors:
            return Response({
                'err': 1,
                'errors': errors,
            })

        transaction = self.transaction

        # lie le conjoint à une nouvelle adresse copie de la première
        adresse_2 = copy.copy(adresse_1)
        adresse_2.id = None
        adresse_2.save()

        assert adresse_2.id and adresse_1.id != adresse_2.id

        c = (Relation.objects.filter(left=conjoint, right=adresse_1, schema__slug=utils.HABITE_REL)
             .update(right=adresse_2, modified=transaction))
        assert c == 1

        # mise à jour de l'adresse des enfants non communs
        c = Relation.objects.exclude(left__in=enfants_communs).filter(
            left__right_relations__left=conjoint,
            right=adresse_1, schema__slug=utils.HABITE_REL).update(right=adresse_2,
                                                                   modified=transaction)

        # crée la nouvelle relation utils.HABITE_REL et pose le flag principale si demandé
        habite_schema = RelationSchema.objects.get(slug=utils.HABITE_REL)
        for enfant in enfants_communs:
            rel_adresse_1 = Relation.objects.get(left=enfant, right=adresse_1)

            rel_adresse_2 = Relation(
                left=enfant, right=adresse_2,
                schema=habite_schema,
                created=transaction,
                content={
                    'principale': False,
                })

            if enfant.id in adresse_principale_1:
                rel_adresse_1.content['principale'] = True
            else:
                rel_adresse_1.content['principale'] = False
                rel_adresse_2.content['principale'] = True
            rel_adresse_1.modified = transaction
            rel_adresse_1.save()
            rel_adresse_2.save()

        # supprime la relation
        relation = Relation.objects.filter(
            left__in=[individu, conjoint],
            right__in=[individu, conjoint],
            schema__slug=utils.UNION_REL).select_related().get()
        relation.delete()

        utils.journalize(
            individu,
            meta=serializer.journal_meta,
            text=u'Déclaration de séparation',
            transaction=transaction,
            conjoint_id=conjoint.id,
            conjoint_text=individu_to_text(conjoint))
        utils.journalize(
            conjoint,
            meta=serializer.journal_meta,
            text=u'Déclaration de séparation',
            transaction=transaction,
            individu_id=individu.id,
            individu_text=individu_to_text(individu))
        fragments.DeclarationSeparation.pour_chaque_application(
            relation,
            meta=serializer.journal_meta,
            transaction=transaction)
        return Response({
            'err': 0,
        })

separation = SeparationView.as_view()


class DeclarationDeDecesSerializer(JournalSerializerMixin):
    date_de_deces = serializers.DateField()


class DeclarationDeDecesView(IndividuViewMixin, TransactionalView):
    def post(self, request, identifier, format=None):
        individu = self.get_individu(identifier)

        if individu.content.get('date_de_deces'):
            return Response({
                'err': 1,
                'errors': [
                    u'cette personne est déjà déclaré décédée.',
                ]
            })

        serializer = DeclarationDeDecesSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)

        transaction = self.transaction
        # on stocke la date du décés
        individu.content['date_de_deces'] = serializer.validated_data['date_de_deces'].isoformat()
        individu.modified = transaction
        individu.save()

        # on supprime les relations
        Relation.objects.filter(schema__slug=utils.RESPONSABILITE_LEGALE_REL,
                                left=individu).delete()
        Relation.objects.filter(schema__slug=utils.RESPONSABILITE_LEGALE_REL,
                                right=individu).delete()
        Relation.objects.filter(schema__slug=utils.UNION_REL, left=individu).delete()
        Relation.objects.filter(schema__slug=utils.UNION_REL, right=individu).delete()

        # si l'adresse est partagée avec un autre adulte on l'individualise
        adresse = utils.adresse(individu)
        if adresse.right_relations.filter(left__content__statut_legal='majeur').count() > 1:
            new_adresse = copy.copy(adresse)
            new_adresse.created = transaction
            new_adresse.id = None
            new_adresse.save()
            assert new_adresse.id and new_adresse.id != adresse.id
            c = Relation.objects.filter(schema__slug=utils.HABITE_REL, left=individu,
                                        right=adresse).update(right=new_adresse)
            assert c == 1
        # si l'adresse n'est partagée qu'avec des enfants, on les débranche
        else:
            Relation.objects.filter(
                schema__slug=utils.HABITE_REL, right=adresse).exclude(
                    left=individu).delete()
        utils.journalize(
            individu,
            meta=serializer.journal_meta,
            text=u'Déclaration de décés',
            transaction=transaction)
        return Response({
            'err': 0,
        })

declaration_de_deces = DeclarationDeDecesView.as_view()


class SuppressionLienDeResponsabiliteView(IndividuViewMixin, TransactionalView):
    def post(self, request, identifier, identifier_enfant, format=None):
        individu = self.get_individu(identifier)
        enfant = self.get_individu(identifier_enfant)

        try:
            relation = Relation.objects.get(left=individu, right=enfant,
                                            schema__slug=utils.RESPONSABILITE_LEGALE_REL)
        except Relation.DoesNotExist:
            return Response({
                'err': 1,
                'errors': [
                    u'cet adulte n\'a pas de responsabilité légale sur cet enfant',
                ]
            }, status=400)

        c = Relation.objects.filter(
            right=enfant,
            schema__slug=utils.RESPONSABILITE_LEGALE_REL).count()
        if c < 2:
            return Response({
                'err': 1,
                'errors': [
                    u'cet enfant n\'a qu\'un responsable légal',
                ]
            }, status=400)

        # supression de la relation
        relation = Relation.objects.filter(
            left=individu, right=enfant,
            schema__slug=utils.RESPONSABILITE_LEGALE_REL).select_related().get()
        relation.delete()

        # suppression éventuelle de l'adresse commune entre l'enfant et l'adulte
        # si elle n'appartient pas aussi à un autre adulte ayant des responsabilités
        # légales sur le même enfant
        adresse = utils.adresse(individu)
        other_individu_ids = [rel.left_id for rel in
                              Relation.objects.filter(left__content__statut_legal='majeur',
                                                      right=adresse,
                                                      schema__slug=utils.HABITE_REL)]
        if not Relation.objects.filter(left_id__in=other_individu_ids,
                                       right=enfant,
                                       schema__slug=utils.RESPONSABILITE_LEGALE_REL).exists():
            Relation.objects.filter(left=enfant, right=adresse).delete()

        serializer = JournalSerializerMixin(data=request.data)
        serializer.is_valid()
        utils.journalize(
            individu,
            meta=serializer.journal_meta,
            text=u'Supression de lien de responsabilité légale',
            statut=relation.content['statut'],
            enfant_id=enfant.id,
            enfant_text=individu_to_text(enfant),
            transaction=self.transaction)
        utils.journalize(
            enfant,
            meta=serializer.journal_meta,
            text=u'Supression de lien de responsabilité légale',
            statut=relation.content['statut'],
            adulte_id=individu.id,
            adulte_text=individu_to_text(individu),
            transaction=self.transaction)
        fragments.SuppressionResponsabiliteEnfant.pour_chaque_application(
            relation,
            meta=serializer.journal_meta,
            transaction=self.transaction)

        return Response({
            'err': 0,
        })

suppression_lien_de_responsabilite = SuppressionLienDeResponsabiliteView.as_view()


class SynchronisationSerializer(JournalSerializerMixin):
    applications = serializers.ListField(
        child=serializers.CharField())
    individus = serializers.ListField(
        child=serializers.IntegerField())


class Synchronisation(TransactionalView):
    def post(self, request, format=None):
        qs = Entity.objects.prefetch_related(
            'left_relations__schema', 'left_relations__right',
            'right_relations__schema', 'right_relations__left',
        ).filter(schema__slug=utils.INDIVIDU_ENT)

        serializer = SynchronisationSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)
        individus = []

        for application in serializer.validated_data['applications']:
            app_dfn = utils.get_application(application)
            if not app_dfn or 'rsu_ws_url' not in app_dfn:
                return Response({
                    'err': 1,
                    'errors': [
                        u'l\'application "%s" est invalide.' % application,
                    ]
                }, status=400)

        for individu_id in serializer.validated_data['individus']:
            try:
                individu = qs.get(id=individu_id)
            except Entity.DoesNotExist:
                return Response({
                    'err': 1,
                    'errors': [
                        u'l\'individu %s est inconnu.' % individu_id,
                    ]
                }, status=400)
            individus.append(individu)

        errors = list(fragments.synchronize(
            serializer.validated_data['applications'],
            individus,
            meta=serializer.journal_meta,
            transaction=self.transaction))

        if errors:
            return Response({
                'err': 1,
                'errors': errors,
            })

        return Response({
            'err': 0,
        })

synchronization = Synchronisation.as_view()


class SagaTiers(APIView):
    def get(self, request, application, identifier, format=None):
        app_dfn = utils.get_application(application)
        if not app_dfn or not app_dfn.get('rsu_ws_url'):
            raise Http404('unknown application')
        qs = Entity.objects.filter(**{'content__cles_de_federation__%s' % application: identifier})
        qs = qs.select_for_update()

        try:
            individu = qs.get()
        except Entity.DoesNotExist:
            raise Http404('unknown identifier')
        if 'saga' not in individu.content['cles_de_federation']:
            individu.content['cles_de_federation']['saga_tiers'] = (
                'RG%013d' % utils.get_next_saga_sequence())
            individu.save()
        return Response({
            'code': individu.content['cles_de_federation']['saga_tiers']
        })

saga_tiers = SagaTiers.as_view()


class SagaMixin(object):
    def get_ws(self):
        app_dfn = utils.get_application('saga')
        if not app_dfn or 'url' not in app_dfn:
            return None, Response({
                'err': 1,
                'errors': [
                    'URL de l\'application SAGA non configurée',
                ]
            })

        ws = self.ws = saga.Saga(app_dfn['url'])
        return ws, None

    def get_factures(self, identifier):
        self.individu = individu = self.get_individu(identifier)

        if 'saga_tiers' not in individu.content['cles_de_federation']:
            return None, Response({
                'err': 1,
                'errors': [
                    'l\'individu n\'a pas de code tiers SAGA',
                ]
            })
        code_tiers = individu.content['cles_de_federation']['saga_tiers']

        ws, error_response = self.get_ws()
        if error_response:
            return None, error_response

        federation = individu.content['cles_de_federation'].get('saga')
        if not federation:
            # obtention d'un verrou sur l'entité avant mise à jour
            individu = Entity.objects.select_for_update().get(id=individu.id)
            federation = individu.content['cles_de_federation'].get('saga')
            if not federation:
                federation, error = ws.resolve_code_tiers(code_tiers)
                if error:
                    return None, Response({
                        'err': 1,
                        'errors': [
                            error,
                        ]
                    })
                individu.content['cles_de_federation']['saga'] = federation
                individu.save()
        factures, error = ws.factures(federation)
        if error:
            return None, Response({
                'err': 1,
                'errors': [
                    error,
                ]
            })

        return factures, None


class SagaFactures(SagaMixin, IndividuViewMixin, APIView):
    def get(self, request, identifier, format=None):
        factures, error_response = self.get_factures(identifier)

        if error_response:
            return error_response

        def to_json(o):
            d = {}
            for field in o._fields:
                if field == 'creances':
                    d['creances'] = [to_json(creance) for creance in o.creances]
                else:
                    d[field] = getattr(o, field)
            return d

        return Response({
            'err': 0,
            'data': [
                to_json(facture) for facture in factures
            ],
        })


saga_factures = SagaFactures.as_view()


class TransactionSagaSerializer(serializers.Serializer):
    num_facture = serializers.CharField()
    urlretour_asynchrone = serializers.URLField()
    urlretour_synchrone = serializers.URLField()
    email = serializers.EmailField(required=False)


class SagaTransaction(SagaFactures):
    def post(self, request, identifier, format=None):
        factures, error_response = self.get_factures(identifier)

        if error_response:
            return error_response

        serializer = TransactionSagaSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)

        for facture in factures:
            if facture.num == serializer.validated_data['num_facture']:
                break
        else:
            return Response({
                'err': 1,
                'errors': [
                    'numéro de facture inconnu',
                ]
            })

        data = serializer.validated_data
        email = data.get('email') or self.individu.content['email'] or ''
        url, error = self.ws.transaction(
            facture=facture,
            urlretour_asynchrone=data['urlretour_asynchrone'],
            urlretour_synchrone=data['urlretour_synchrone'],
            email=email)
        if error:
            return Response({
                'err': 1,
                'errors': [
                    error,
                ]
            })

        return Response({
            'err': 0,
            'data': {
                'url': url,
            },
        })

saga_transaction = SagaTransaction.as_view()


class RetourSagaSerializer(serializers.Serializer):
    idop = serializers.CharField()


class SagaRetourAsynchrone(SagaMixin, APIView):
    def post(self, request, format=None):
        serializer = RetourSagaSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)

        ws, error_response = self.get_ws()
        if error_response:
            return error_response

        result, error = ws.page_retour_asynchrone(serializer.validated_data['idop'])
        if error:
            return Response({
                'err': 1,
                'errors': [
                    error,
                ]
            })

        return Response({
            'err': 0,
            'data': result,
        })

saga_retour_asynchrone = SagaRetourAsynchrone.as_view()


class SagaRetourSynchrone(SagaMixin, APIView):
    def post(self, request, format=None):
        serializer = RetourSagaSerializer(data=request.data)
        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)

        ws, error_response = self.get_ws()
        if error_response:
            return error_response

        result, error = ws.page_retour_synchrone(serializer.validated_data['idop'])
        if error:
            return Response({
                'err': 1,
                'errors': [
                    error,
                ]
            })

        return Response({
            'err': 0,
            'data': result,
        })

saga_retour_synchrone = SagaRetourSynchrone.as_view()


class QFLireQuotientsValide(APIView):
    def get(self, request, format=None):
        ws = qf.QF()
        response, error = ws.lire_quotients_valides(now())
        if error:
            return Response({
                'err': 1,
                'errors': [error],
            })
        return Response({
            'err': 0,
            'data': [
                {
                    'id': data['nature-qf'],
                    'text': data['libelle'],
                } for data in response
            ]
        })

qf_lire_quotiens_valides = QFLireQuotientsValide.as_view()


class QFSimuleSerializer(serializers.Serializer):
    nature_qf = serializers.IntegerField()
    annee_imposition = serializers.IntegerField()
    rfr = serializers.IntegerField()
    nb_parts = serializers.IntegerField()
    monoparentalite = serializers.BooleanField()
    annee_imposition_concubin = serializers.IntegerField(required=False)
    rfr_concubin = serializers.IntegerField(required=False)
    nb_parts_concubin = serializers.IntegerField(required=False)


class QFSimuler(APIView):
    def post(self, request, format=None):
        serializer = QFSimuleSerializer(data=request.data)

        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)
        ws = qf.QF()
        d = serializer.validated_data
        result, error = ws.simuler_qf(**d)
        if error:
            return Response({
                'err': 1,
                'errors': [error],
            })
        return Response({
            'err': 0,
            'data': result,
        })

qf_simuler = QFSimuler.as_view()


class QFCalculer(IndividuViewMixin, APIView):
    def post(self, request, identifier, format=None):
        serializer = QFSimuleSerializer(data=request.data)

        if not serializer.is_valid():
            return Response({
                'err': 1,
                'errors': flatten_errors(serializer.errors),
            }, status=400)
        d = serializer.validated_data.copy()
        individu = self.get_individu(identifier)
        d['individu'] = individu
        ws = qf.QF()
        result, error = ws.calcul_qf(**d)
        if error:
            return Response({
                'err': 1,
                'errors': [error],
            })
        return Response({
            'err': 0,
            'data': result,
        })

    def get(self, request, identifier, format=None):
        individu = self.get_individu(identifier)
        ws = qf.QF()
        result, error = ws.lire_quotient_familial(individu, now())
        if error:
            return Response({
                'err': 1,
                'errors': [error],
            })
        return Response({
            'err': 0,
            'data': result,
        })

qf_calculer = QFCalculer.as_view()


class QFEditerCarte(IndividuViewMixin, APIView):
    def get(self, request, identifier, id_qf, format=None):
        individu = self.get_individu(identifier)
        ws = qf.QF()
        result, error = ws.editer_carte(individu, int(id_qf))
        if error:
            return Response({
                'err': 1,
                'errors': [error],
            })
        response = HttpResponse(result, content_type='application/pdf')
        response['Content-Disposition'] = 'attachment; filename="qf-%s-%s.pdf"' % (
            id_qf, now().isoformat())
        return response

qf_editer_carte = QFEditerCarte.as_view()
